/****************************************************************************
**
** Copyright (C) 2006-2009 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
** 
** This file may be used under the terms of the GNU General Public License
** version 3 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "codedefinition.h"

/*!
	\file codedefinition.cpp
	\brief Implementation of CodeDefinition
	
	\see CodeDefinition
*/

/*!
	\ingroup language
	@{
	
	\class CodeDefinition
	\brief Implementation of language definition.
	
*/

#include "automaton.h"
#include "generator.h"

#include "qlanguagefactory.h"

#include "qdocument.h"
#include "qdocument_p.h"
#include "qdocumentline.h"
#include "qdocumentcursor.h"
#include "qdocumentbuffer.h"

#include <QFile>
#include <QDebug>
#include <QKeyEvent>
#include <QDataStream>

/*!
	\brief Empty constructor
*/
CodeDefinition::CodeDefinition(HashedAutomaton *a)
 : m_lexer(a)
{
}

/*!
	\brief Empty destructor
*/
CodeDefinition::~CodeDefinition()
{
}

void CodeDefinition::load(const QString& file, QLanguageFactory::LangData *d, QFormatScheme *s)
{
	//Generator::TokenMap tm;
	//HashedAutomaton *a = Generator().createLexer(file, tm);
	
	QFile f(file);
	f.open(QFile::ReadOnly);
	QDataStream stream(&f);
	
	HashedAutomaton *a = HashedAutomaton::deserialize(stream);
	
	// TODO : formatting/context registration and such
	CodeDefinition *cd = new CodeDefinition(a);
	cd->registerContext(a);
	
	d->d = cd;
	d->e = 0;
	d->s = s;
}

void CodeDefinition::registerContext(HashedAutomaton *a)
{
	QHash<HashedAutomaton*, int>::const_iterator it = m_contextIds.constFind(a);
	
	if ( it == m_contextIds.constEnd() )
	{
		m_contextIds[a] = m_contexts.count();
		m_contexts << a;
		
		foreach ( HashedAutomaton *c, a->sub )
			registerContext(c);
	}
}

/*!
	\return The language supported by this definition
*/
QString CodeDefinition::language() const
{
	return m_lang;
}

/*!
	\return the file extensions corrseponding to the supported language
	
	\see language()
	\see QFileInfo::completeSuffix()
*/
QStringList CodeDefinition::extensions() const
{
	return m_exts;
}

inline bool move(const StateHash& sh, quint16 c, StateId& d)
{
	StateHash::const_iterator it = sh.constFind(c);
	
	if ( it != sh.constEnd() )
	{
		d = *it;
		return true;
	}
	
	d = 0;
	return false;
}

/*!
	\brief Entry point for syntax highlighting
*/
int CodeDefinition::tokenize(QDocument *d, int firstLine, int count)
{
	int lastLine = firstLine + count;
	QDocumentBuffer::iterator it = d->impl()->find(firstLine);
	
	// build proper context stack to continue parsing at start of edit range
	QStack<int> cxtidx;
	HashedAutomaton *context;
	QStack<HashedAutomaton*> cxt;
	
	cxt.append(m_lexer);
	
	if ( firstLine > 0 )
	{
		QDocumentBuffer::iterator tmp = it;
		tmp.moveBy(-1);
		
		const QVector<int>& cxts = tmp.lineHandle()->contexts();
		
		//qDebug() << "context stack : " << cxts << endl;
		
		for ( int i = 0; i + 2 < cxts.count(); i += 3 )
		{
			if ( cxts.at(i + 1) == -1 )
			{
				// TODO : take language field into account, embedding and such...
				int cxtId = cxts.at(i + 2) & ContextIdMask;
				cxt.append(m_contexts[cxtId]);
				//qDebug(">>cxtId : %i", cxtId);
			}
		}
	}
	
	context = cxt.pop();
	
	//
	
	QChar c, tc;
	StateId s = 0, tmp = 0;
	int offset, idx = 0, length;
	const StateHash *sha = context->states.constData();
	
	static const QLatin1Char eol('\n');
	
	int m_line = firstLine, m_column = 0;
	
	// lexing loop, line by line
	while ( !it.atEnd() )
	{
		QDocumentLineHandle *h = it.lineHandle();
		
		const QChar *d = h->text().constData();
		
		idx = m_column;
		length = h->length();
		
		QVector<int> oldCxtStack, newCxtStack;
		
		if ( m_line + 1 >= lastLine )
		{
			const QVector<int>& cxts = h->contexts();
			
			for ( int i = 0; i + 2 < cxts.count(); i += 3 )
			{
				if ( cxts.at(i) == -1 || cxts.at(i + 1) == -1 )
				{
					int cxtId = cxts.at(i + 2);
					// TODO : take language field into account, embedding and such...
					oldCxtStack.append(cxtId);
				}
			}
		}
		
		// TODO : selective clearing if we really want to support
		// fine-grained hl update (and possibly extend the scope
		// if overlaps with an existing token...
		h->clearTokens();
		h->clearContexts();
		
		QMap<int, QPair<int, int> > cxtData;
		
		// actual lexing
		while ( idx <= length )
		{
			s = tmp = 0;
			c = idx != length ? *d : eol;
			
			if ( move(sha[s], c.unicode(), tmp) )
			{
				tc = c;
				offset = idx;
				
				quint16 tcu;
				const QChar *td = d;
				int col = m_column;
				
				do
				{
					s = tmp;
					++td;
					++offset;
					
					if ( offset > length )
						break;
					
					if ( offset == length )
					{
						tcu = '\n';
					} else {
						tcu = td->unicode();
					}
					
					++col;
					
				} while ( move(sha[s], tcu, tmp) );
				
				offset -= idx;
				
				//qDebug("?production : %s [%i]", qPrintable(QString(d, offset)), s);
				
				if ( !s )
				{
					++d;
					++idx;
					++m_column;
					continue;
				}
				
				QHash<StateId, Match>::const_iterator it = context->match.constFind(s);
				
				if ( it != context->match.constEnd() )
				{
					int endColumn = col;
					
					//qDebug("production : %s [%i]", qPrintable(QString(d, offset)), s);
					
					Match m = *it;
					quint32 mt = qMatchType(m);
					
					if ( mt & ::Context )
					{
						//out->consume(tok);
						//qDebug("context");
						
						if ( mt & Begin )
						{
							cxt.push(context);
							context = context->sub.at(qMatchJump(m));
							
							if ( context->flags != Merge )
							{
								if ( context->flags & Hide )
								{
									
								} else {
									
								}
								
								//enter(m_line, m_column, context->format, &ud);
							}
							
							cxtidx.push(idx);
							//qDebug("entering context %s", qPrintable(context->id));
						} else if ( cxt.count() ) {
							//qDebug("leaving context %s", qPrintable(context->id));
							int last;
							
							if ( context->flags == Merge )
							{
								if ( cxtidx.count() )
								{
									last = cxtidx.pop();
								} else {
									// merge a context not started on that line
									// the whole line is part of it then...
									// what of possible subcontexts/matches?
									// => they are screwed and that's not our problem
									last = 0;
									
								}
								//int last = cxtidx.pop();
								//tok.id = context->format;
								//tok.production = QStringRef(&m_data, last, idx + offset - last);
								h->addToken(last, qMin(endColumn, length) - last, context->format);
								//qDebug("hl:token(%i, %i, %i)", last, qMin(endColumn, length) - last, context->format);
							} else {
								//leave(m_line, endColumn, context->format, &ud);
								int len = qMin(endColumn, length);
								const int cxtId = m_contextIds.value(context);
								
								if ( cxtidx.count() )
								{
									last = cxtidx.pop();
									len -= last;
								} else {
									last = -1;
								}
								
								//h->addContext(last, endColumn - last, context->format);
								cxtData.insertMulti(last, qMakePair(len, cxtId));
							}
							
							context = cxt.pop();
						} else {
							qWarning("broken context nest.");
						}
						
						sha = context->states.constData();
					} else { //if ( context->flags != Merge ) {
						//qDebug("hl:token(%i, %i, %i)", m_column, endColumn - m_column, qMatchFormat(m));
						
						h->addToken(m_column, qMin(endColumn, length) - m_column, qMatchFormat(m));
					}
					
					// tell the source that we consumed some chars
					if ( !(mt & NoConsume) )
					{
						idx += offset;
						d += offset;
					}
					
					m_column = endColumn;
					continue;
				}
			} else {
				//qDebug("%c sterile", c.toLatin1());
			}
			
			++m_column;
			++idx;
			++d;
		}
		
		
		++m_line;
		// commit contexts in proper order
		int n = cxt.count(), nn = cxtidx.count();
		
		int anchor = nn ? cxtidx.last() : -1;
		
		if ( nn )
			--nn;
		
		int i = 1;
		
		//qDebug("n:%i, nn:%i", n, nn);
		
		for ( ; i < (n - nn); ++i )
		{
			int cxtId = m_contextIds.value(cxt.at(i))  | languageId();
			
			//qDebug(">][ : %i", cxtId);
			h->addContext(-1, -1, cxtId);
			
			if ( m_line >= lastLine )
			{
				newCxtStack.append(cxtId);
			}
		}
		
		for ( int j = i; j < n; ++j )
		{
			int cxtId = m_contextIds.value(cxt.at(i))  | languageId();
			cxtData[cxtidx.at(j - i)] = qMakePair(-1, cxtId);
			
			if ( m_line >= lastLine )
			{
				newCxtStack.append(cxtId);
			}
		}
		
		if ( cxt.count() )
		{
			int cxtId = m_contextIds.value(context) | languageId();
			
			if ( m_line >= lastLine )
				newCxtStack.append(cxtId);
			
			if ( anchor != -1 )
			{
				cxtData[anchor] = qMakePair(-1, cxtId);
			} else {
				//qDebug(">][ : %i (top)", cxtId);
				h->addContext(-1, -1, cxtId);
			}
		}
		
		QMap<int, QPair<int, int> >::const_iterator cit = cxtData.constBegin();
		
		while ( cit != cxtData.constEnd() )
		{
			//qDebug(">[%i,%i? : %i", cit.key(), cit->first, cit->second);
			h->addContext(cit.key(), cit->first, cit->second);
			
			++cit;
		}
		
		cxtidx.clear();
		it.moveBy(1);
		m_column = 0;
		
		if ( m_line >= lastLine )
		{
			// check whether further lexing is required
			//qWarning("TODO : continuation");
			
			//qDebug() << oldCxtStack << " vs " << newCxtStack << endl;
			
			if ( newCxtStack == oldCxtStack )
				break;
			
			//qDebug("continuing");
			newCxtStack.clear();
			oldCxtStack.clear();
			//break;
		}
	}
	
	// TODO : extend if needed...
	return m_line - firstLine;
}

/*!
	\brief Return the string starting a single line comment, if any offered by the language
*/
QString CodeDefinition::singleLineComment() const
{
	return QString();
}

/*!
	\brief Let language specify which line mark should be toggled by left clicking a line mark panel
*/
QString CodeDefinition::defaultLineMark() const
{
	return QString();
}

/*!
	\brief Brace matching entry point
*/
void CodeDefinition::clearMatches(QDocument *d)
{
	Q_UNUSED(d)
}

/*!
	\brief Brace matching entry point
*/
void CodeDefinition::match(QDocumentCursor& c)
{
	Q_UNUSED(c)
}

/*!
	\brief Return the indent to use when inserting a line at a given cursor position
*/
QString CodeDefinition::indent(const QDocumentCursor& c)
{
	Q_UNUSED(c)
	
	return QString();
}

/*!
	\brief Determines whether the given key event at the given position should cause unindent to happen
*/
bool CodeDefinition::unindent (const QDocumentCursor& c, const QString& ktxt)
{
	Q_UNUSED(c)
	Q_UNUSED(ktxt)
	
	return false;
}

/*!
	\brief Expand a collapsed block at a given line
*/
void CodeDefinition::expand(QDocument *d, int line)
{
	Q_UNUSED(d)
	Q_UNUSED(line)
}

/*!
	\brief Collapse a text block at a given line
*/
void CodeDefinition::collapse(QDocument *d, int line)
{
	Q_UNUSED(d)
	Q_UNUSED(line)
}

/*!
	\brief Compute the collapse state of a line
*/
int CodeDefinition::blockFlags(QDocument *d, int line, int depth) const
{
	Q_UNUSED(d)
	Q_UNUSED(line)
	Q_UNUSED(depth)
	
	return 0;
}

/*! @} */
